from ..t import In, Out
from ..circuit import Circuit
from ..wire import wire


def build_io_args(annotations):
    io_args = {}

    for param, annotation in annotations.items():
        if param == "return":
            if isinstance(annotation, tuple):
                for i, elem in enumerate(annotation):
                    io_args[f"O{i}"] = Out(elem)
            else:
                io_args["O"] = Out(annotation)
            continue
        annotation = In(annotation)
        io_args[param] = annotation
    return io_args


def build_call_args(io, annotations):
    return [getattr(io, param) for param in annotations if param != "return"]


def wire_call_result(io, call_result, annotations):
    if isinstance(call_result, Circuit):
        if not len(call_result.interface.outputs()) == 1:
            raise TypeError(
                "Expected register return instance with one output")
        call_result = call_result.interface.outputs()[0]
    if isinstance(annotations["return"], tuple):
        for i in range(len(annotations["return"])):
            wire(getattr(io, f"O{i}"), call_result[i])
    else:
        wire(io.O, call_result)



import functools
import ast
import inspect
import textwrap
from collections import OrderedDict
from magma.logging import root_logger
from magma.backend.util import make_relative
import astor
import os
import traceback
import magma.ast_utils as ast_utils
import types
from magma.debug import debug_info
from magma.ssa import convert_tree_to_ssa
from magma.config import get_debug_mode
from ..t import Type
from ..protocol_type import MagmaProtocol
import itertools
import typing
from ast_tools.stack import _SKIP_FRAME_DEBUG_STMT, SymbolTable
import magma as m


_logger = root_logger()


class CircuitDefinitionSyntaxError(Exception):
    pass


def m_dot(attr):
    return ast.Attribute(ast.Name("m", ast.Load()), attr, ast.Load())


def report_transformer_warning(message, filename, lineno, line):
    warning(f"\033[1m{make_relative(filename)}:{lineno}: {message}")
    warning(line)


class IfTransformer(ast.NodeTransformer):
    def __init__(self, filename, lines):
        super().__init__()
        self.filename = filename
        if lines:
            self.lines, self.starting_line = lines
        else:
            self.lines, self.starting_line = None, None

    def flatten(self, _list):
        """1-deep flatten"""
        flat_list = []
        for item in _list:
            if isinstance(item, list):
                flat_list.extend(item)
            else:
                flat_list.append(item)
        return flat_list

    def visit_If(self, node):
        # Flatten in case there's a nest If statement that returns a list
        node.body = self.flatten(map(self.visit, node.body))
        if not hasattr(node, "orelse"):
            raise NotImplementedError("If without else")
        node.orelse = self.flatten(map(self.visit, node.orelse))
        seen = OrderedDict()
        for stmt in node.body:
            if not isinstance(stmt, ast.Assign):
                # TODO: Print info from original source file/line
                raise CircuitDefinitionSyntaxError(
                    f"Expected only assignment statements in if statement, got"
                    f" {type(stmt)}")
            if len(stmt.targets) > 1:
                raise NotImplementedError("Assigning more than one value")
            key = ast.dump(stmt.targets[0])
            if key in seen:
                if self.filename:
                    # TODO: Print the line number
                    report_transformer_warning(
                        "Assigning to value twice inside `if` block,"
                        " taking the last value (first value is ignored)",
                        self.filename, node.lineno + self.starting_line,
                        self.lines[node.lineno])
                else:
                    warning(
                        "Assigning to value twice inside `if` block,"
                        " taking the last value (first value is ignored)")
            seen[key] = stmt
        orelse_seen = set()
        for stmt in node.orelse:
            key = ast.dump(stmt.targets[0])
            if key in seen:
                if key in orelse_seen:
                    if self.filename:
                        report_transformer_warning(
                            "Assigning to value twice inside `else` block,"
                            " taking the last value (first value is ignored)",
                            self.filename, node.lineno + self.starting_line,
                            self.lines[node.lineno])
                    else:
                        warning(
                            "Assigning to value twice inside `else` block,"
                            " taking the last value (first value is ignored)")
                orelse_seen.add(key)
                seen[key].value = ast.Call(
                    ast.Name("phi", ast.Load()),
                    [ast.List([stmt.value, seen[key].value],
                              ast.Load()), node.test],
                    [])
            else:
                if self.filename:
                    report_transformer_warning(
                        "NOT IMPLEMENTED: Assigning to a variable once in"
                        " `else` block (not in then block)",
                        self.filename, node.lineno + self.starting_line,
                        self.lines[node.lineno])
                else:
                    warning(
                        "NOT IMPLEMENTED: Assigning to a variable once in"
                        " `else` block (not in then block)")
                raise NotImplementedError()
        return [node for node in seen.values()]

    def visit_IfExp(self, node):
        if not hasattr(node, "orelse"):
            raise NotImplementedError("If without else")
        node.body = self.visit(node.body)
        node.orelse = self.visit(node.orelse)
        return ast.Call(
            ast.Name("phi", ast.Load()),
            [ast.List([node.orelse, node.body],
                      ast.Load()), node.test],
            [])


class FunctionToCircuitDefTransformer(ast.NodeTransformer):
    def __init__(self, renamed_args):
        super().__init__()
        self.IO = {}
        self.renamed_args = renamed_args

    def visit(self, node):
        new_node = super().visit(node)
        if new_node is not node:
            return ast.copy_location(new_node, node)
        return node

    def qualify(self, node, direction):
        return ast.Call(m_dot(direction), [node], [])

    def visit_FunctionDef(self, node):
        names = self.renamed_args
        types = [arg.annotation for arg in node.args.args]
        keywords = []
        for name, type_ in zip(names, types):
            self.IO[name + "_0"] = name  # Add ssa rename
            keywords.append(ast.keyword(name, self.qualify(type_, "In")))
        if isinstance(node.returns, ast.Tuple):
            for i, elt in enumerate(node.returns.elts):
                keywords.append(ast.keyword(f"O{i}", self.qualify(elt, "Out")))
        else:
            keywords.append(ast.keyword("O", self.qualify(node.returns, "Out")))
        IO = ast.Call(m_dot("IO"), [], keywords)
        node.body = [self.visit(s) for s in node.body]
        if isinstance(node.returns, ast.Tuple):
            for i, elt in enumerate(node.returns.elts):
                node.body.append(ast.Expr(ast.Call(
                    m_dot("wire"),
                    [ast.Name(f"O{i}", ast.Load()),
                     ast.Attribute(ast.Name("io", ast.Load()), f"O{i}",
                                   ast.Load())],
                    []
                )))
        else:
            node.body.append(ast.Expr(ast.Call(
                m_dot("wire"),
                [ast.Name("O", ast.Load()),
                 ast.Attribute(ast.Name("io", ast.Load()), "O", ast.Load())],
                []
            )))
        # class {node.name}(m.Circuit):
        #     IO = {IO}
        #     @classmethod
        #     def definition(io):
        #         {body}
        class_def = ast.ClassDef(
            node.name,
            [ast.Attribute(ast.Name("m", ast.Load()), "Circuit", ast.Load())],
            [], [ast.Assign([ast.Name("io", ast.Store())], IO)] + node.body,
            [])
        return class_def

    def visit_Name(self, node):
        if node.id in self.IO and isinstance(node.ctx, ast.Load):
            return ast.Attribute(ast.Name("io", ast.Load()), self.IO[node.id],
                                 ast.Load())
        return node

#     def visit_Return(self, node):
#         node.value = self.visit(node.value)
#         if isinstance(node.value, ast.Tuple):
#             return ast.Assign(
#                 [ast.Tuple([ast.Name(f"O{i}", ast.Store())
#                  for i in range(len(node.value.elts))], ast.Store())],
#                 node.value
#             )
#         return ast.Assign([ast.Name("O", ast.Store())], node.value)


def _combinational(defn_env: dict, fn: types.FunctionType):
    tree = ast_utils.get_func_ast(fn)
    tree, renamed_args = convert_tree_to_ssa(tree, defn_env)
    tree = FunctionToCircuitDefTransformer(renamed_args).visit(tree)
    tree = ast.fix_missing_locations(tree)
    filename = None
    lines = None
    if get_debug_mode():
        filename = inspect.getsourcefile(fn)
        lines = inspect.getsourcelines(fn)
    tree = IfTransformer(filename, lines).visit(tree)
    tree = ast.fix_missing_locations(tree)
    tree.decorator_list = ast_utils.filter_decorator(
        combinational, tree.decorator_list, defn_env)
    if "m" not in defn_env:
        defn_env["m"] = m
    if "phi" not in defn_env:
        defn_env["phi"] = m.primitives.mux
    source = "\n"
    for i, line in enumerate(astor.to_source(tree).splitlines()):
        source += f"    {i}: {line}\n"

    _logger.debug(source)
    circuit_def = ast_utils.compile_function_to_file(tree, fn.__name__,
                                                     defn_env)
    if get_debug_mode() and getattr(circuit_def, "debug_info", False):
        circuit_def.debug_info = debug_info(circuit_def.debug_info.filename,
                                            circuit_def.debug_info.lineno,
                                            inspect.getmodule(fn))

    @functools.wraps(fn)
    def func(*args, **kwargs):
        if (len(args) + len(kwargs) and
            not any(isinstance(x, (Type, MagmaProtocol, m.AnonymousCircuitType)) for x in args +
                    tuple(kwargs.values()))):
                # If not called with at least one magma value, use the Python
                # implementation
                return fn(*args, **kwargs)
        return circuit_def()(*args, **kwargs)
    func.__name__ = fn.__name__
    func.__qualname__ = fn.__name__
    # Provide a mechanism for accessing the underlying circuit definition
    setattr(func, "circuit_definition", circuit_def)
    return func

def combinational(
        fn: typing.Callable = None,
        *,
        decorators: typing.Optional[typing.Sequence[typing.Callable]] = None,
        env: SymbolTable = None
        ):
    exec(_SKIP_FRAME_DEBUG_STMT)
    if decorators is not None:
        assert fn is None
        def wrapped(fn):
            exec(_SKIP_FRAME_DEBUG_STMT)
            nonlocal decorators
            decorators = list(itertools.chain(decorators, [wrapped]))
            wrapped_combinational = ast_utils.inspect_enclosing_env(
                    _combinational,
                    decorators=decorators,
                    st=env
            )
            return wrapped_combinational(fn)
        return wrapped

    else:
        wrapped_combinational = ast_utils.inspect_enclosing_env(
                _combinational,
                decorators=[combinational],
                st=env
        )
        return wrapped_combinational(fn)



import hwtypes as ht
from hwtypes import BitVector, UIntVector, SIntVector
from magma.array import Array
from magma.bit import Bit
from magma.bits import Bits, UInt, SInt
from magma.bitutils import clog2, seq2int
from magma.circuit import coreir_port_mapping
from magma.generator import Generator2
from magma.interface import IO
from magma.protocol_type import MagmaProtocol, magma_type
from magma.t import Type, In, Out, Direction
from magma.tuple import Product, Tuple
from magma.conversions import tuple_


class CoreIRCommonLibMuxN(Generator2):
    def __init__(self, N: int, width: int):
        self.name = f"coreir_commonlib_mux{N}x{width}"
        FlatT = Array[width, Bit]
        MuxInT = Product.from_fields("anon", dict(data=Array[N, FlatT],
                                                  sel=Array[clog2(N), Bit]))
        self.io = IO(I=In(MuxInT), O=Out(Array[width, Bit]))
        self.renamed_ports = coreir_port_mapping
        self.coreir_name = "muxn"
        self.coreir_lib = "commonlib"
        self.coreir_genargs = {"width": width, "N": N}
        self.primitive = True
        self.stateful = False

        def simulate(self, value_store, state_store):
            sel = BitVector[clog2(N)](value_store.get_value(self.I.sel))
            out = BitVector[width](value_store.get_value(self.I.data[int(sel)]))
            value_store.set_value(self.O, out)

        self.simulate = simulate


class Mux(Generator2):
    def __init__(self, height: int, T: Type):
        if issubclass(T, BitVector):
            T = Bits[len(T)]
        if issubclass(T, (bool, ht.Bit)):
            T = Bit
        if issubclass(T, MagmaProtocol):
            T = T._to_magma_()
        T = magma_type(T)
        T_str = str(T).replace("(", "").replace(")", "")\
                      .replace(",", "_").replace("=", "_")\
                      .replace("[", "").replace("]", "").replace(" ", "")
        self.name = f"Mux{height}x{T_str}"
        # TODO: Type must be hashable so we can cache.
        N = T.flat_length()

        io_dict = {}
        for i in range(height):
            io_dict[f"I{i}"] = In(T)
        if height == 2:
            io_dict["S"] = In(Bit)
        else:
            io_dict["S"] = In(Array[clog2(height), Bit])
        io_dict["O"] = Out(T)

        self.io = IO(**io_dict)
        mux = CoreIRCommonLibMuxN(height, N)()
        data = [Array[N, Bit](getattr(self.io, f"I{i}").flatten())
                for i in range(height)]
        mux.I.data @= Array[height, Array[N, Bit]](data)
        if height == 2:
            mux.I.sel[0] @= self.io.S
        else:
            mux.I.sel @= self.io.S
        self.io.O @= T.unflatten(mux.O.ts)


def _infer_mux_type(args):
    """
    Try to infer type by traversing arguments in order:
    * If we encounter a magma Type/Protocol, use that
    * BitVector/Bit/bool are converted to their magma equivalent Bits/Bit
    * Python tuple is converted to m.Tuple (note this will invoke m.tuple_ on
      all the arguments, which may raise an error if the tuple arguments are
      not well formed)

    Note that we do not infer from standard python int arguments because we
    cannot, in general, determine the correct bit width (use BitVector instead)
    """
    T = None
    for arg in args:
        if isinstance(arg, (Type, MagmaProtocol)):
            next_T = type(arg).qualify(Direction.Undirected)
        elif isinstance(arg, UIntVector):
            next_T = UInt[len(arg)]
        elif isinstance(arg, SIntVector):
            next_T = SInt[len(arg)]
        elif isinstance(arg, BitVector):
            next_T = Bits[len(arg)]
        elif isinstance(arg, (ht.Bit, bool)):
            next_T = Bit
        elif isinstance(arg, tuple):
            next_T = type(tuple_(arg))
        elif isinstance(arg, int):
            # Cannot infer type without width, use wiring implicit coercion to
            # handle (or raise type error there)
            continue

        if T is not None:
            if issubclass(T, next_T):
                # upcast
                T = next_T
            elif not next_T.is_wireable(T) and not T.is_wireable(next_T):
                raise TypeError(
                    f"Found incompatible types {next_T} and {T} in mux"
                    " inference"
                )
        else:
            T = next_T
    if T is None:
        raise TypeError(
            f"Could not infer mux type from {args}\n"
            "Need at least one magma value, BitVector, bool or tuple")
    if issubclass(T, Tuple):
        args = [tuple_(a) for a in args]
    return T, args


def mux(I: list, S, **kwargs):
    """
    How type inference works on I:
        This operator will traverse the list of inputs I and use the first
        magma value to determine the type.  This allows I to contain coerceable
        objects like ints (e.g. `mux([1, x], s)` where `x: UInt[2]`).  Coercion
        is peformed when wiring the arguments to the input of an instance of
        Mux(T) (where T is the type of the first magma value) and will raise an
        error there if one of the values cannot be coerced to T.
        **NOTE** This will fail if the type of the first magma value cannot
        coerce subsequent arguments (even though the type of a later argument
        might be able to coerce the type of the earliest argument).  We plan to
        improve the algorithm to support the more general cases, but until
        then, if you run into this problem, use `Mux(T)` where `T` can coerce
        all arguments.
    """
    if isinstance(S, Type) and S.const():
        S = seq2int(S.bits())
    if isinstance(S, int):
        return I[S]
    T, I = _infer_mux_type(I)
    inst = Mux(len(I), T, **kwargs)()
    if len(I) == 2 and isinstance(S, Bits[1]):
        S = S[0]
    result = inst(*I, S)
    for i in range(len(I)):
        if getattr(inst, f"I{i}").value() is None:
            arg = I[i]
            raise TypeError(f"mux arg I[{i}] ({arg}: {type(arg)}) does not "
                            f"match inferred input port type {T}")
    return result


# Monkey patch for ite impl without circular dependency
Bit._mux = staticmethod(mux)


# NOTE(rsetaluri): We monkeypatch this function on to Array due to the circular
# dependency between Mux and Array. See the discussion on
# https://github.com/phanrahan/magma/pull/658.
def _dynamic_mux_select(this, key):
    return mux(this.ts, key)


Array.dynamic_mux_select = _dynamic_mux_select


def dict_lookup(dict_, select, default=0):
    """
    Use `select` as an index into `dict` (similar to a case statement)

    `default` is used when `select` does not match any of the keys and has a
    default value of 0
    """
    output = default
    for key, value in dict_.items():
        output = mux([output, value], key == select)
    return output


def list_lookup(list_, select, default=0):
    """
    Use `select` as an index into `list` (similar to a case statement)

    `default` is used when `select` does not match any of the indices (e.g.
    when the select width is longer than the list) and has a default value of
    0.
    """
    output = default
    for i, elem in enumerate(list_):
        output = mux([output, elem], i == select)
    return output



import functools
from collections import OrderedDict
from collections.abc import Sequence, Mapping
from .compatibility import IntegerTypes
from .t import Type, In, Out, InOut, Direction
from .digital import Digital
from .bit import Bit
from .bits import Bits
from .digital import DigitalMeta
from .clock import Clock, Reset, AsyncReset, AsyncResetN, Enable
from .array import Array
from .bits import Bits, UInt, SInt
from .bfloat import BFloat
from .digital import Digital
from .tuple import Tuple, Product
from .protocol_type import magma_type, magma_value
from .bitutils import int2seq
import hwtypes as ht
from magma.common import is_int


__all__ = ['bit']
__all__ += ['clock', 'reset', 'enable', 'asyncreset', 'asyncresetn']

__all__ += ['array']
__all__ += ['bits', 'uint', 'sint']

__all__ += ['tuple_', 'namedtuple']

__all__ += ['concat', 'repeat']
__all__ += ['sext', 'zext', 'sext_to', 'sext_by', 'zext_to', 'zext_by']
__all__ += ['replace']
__all__ += ['as_bits', 'from_bits']

def can_convert_to_bit(value):
    return isinstance(magma_value(value), (Digital, Array, Tuple, IntegerTypes))


def can_convert_to_bit_type(value):
    return issubclass(magma_type(value), (Digital, Array, Tuple))


def convertbit(value, totype):
    value = magma_value(value)
    if isinstance(value, totype):
        return value

    if not can_convert_to_bit(value):
        raise ValueError(
            "bit can only be used on a Bit, an Array, or an int"
            f"; not {type(value)}")

    if isinstance(value, (Array, Tuple)):
        if len(value) != 1:
            raise ValueError(
                "bit can only be used on arrays and tuples of length 1"
                f"; not {type(value)}")
        value = value[0]
        if not isinstance(value, Digital):
            raise ValueError(
                "bit can only be used on arrays and tuples of bits"
                f"; not {type(value)}")

    assert isinstance(value, (IntegerTypes, Digital))

    if isinstance(value, IntegerTypes):
        # Just return VCC or GND here, otherwise we lose VCC/GND singleton
        # invariant
        return totype(1) if value else totype(0)

    if value.is_input():
        b = In(totype)(name=value.name)
    elif value.is_output():
        b = Out(totype)(name=value.name)
    else:
        b = totype()
    b._wire = value._wire
    return b


def bit(value):
    return convertbit(value, Bit)


def clock(value):
    return convertbit(value, Clock)


def reset(value):
    return convertbit(value, Reset)


def asyncreset(value):
    return convertbit(value, AsyncReset)


def asyncresetn(value):
    return convertbit(value, AsyncResetN)


def enable(value):
    return convertbit(value, Enable)


def convertbits(value, n, totype, checkbit):
    if isinstance(value, totype):
        if n is not None and n != len(value):
            raise ValueError("converting a value should not change the size, use concat, zext, or sext instead.")
        return value

    value = magma_value(value)
    convertible_types = (Digital, Tuple, Array, IntegerTypes,
                         Sequence, ht.BitVector)
    if not isinstance(value, convertible_types):
        raise ValueError(
            "bits can only be used on a Bit, an Array, a Tuple, an int, or a"
            f" Sequence; not : {type(value)}"
        )

    if isinstance(value, IntegerTypes):
        if n is None:
            n = max(value.bit_length(), 1)
        ts = int2seq(value, n)
    elif isinstance(value, ht.BitVector):
        ts = value.bits()
    elif isinstance(value, Sequence):
        ts = list(value)
    elif isinstance(value, Digital):
        if n is None:
            ts = [value]
        elif issubclass(totype, SInt):
            # sext
            ts = n * [value]
        else:
            # zext
            ts = [value] + [Bit(0) for _ in range(n - 1)]
    else:
        ts = [value[i] for i in range(len(value))]

    # create list of types
    Ts = []
    for t in ts:
        T = type(t)
        if T in IntegerTypes:
            T = Out(Bit)
        Ts.append(T)

    convert_to_bit = False
    # check that they are all the same
    for t in Ts:
        if checkbit:
            if not isinstance(magma_type(t), DigitalMeta):
                raise ValueError(
                    "bits can only be used on Arrays or Tuples containing bits"
                    f", not : {t}")
        if t != T:
            if can_convert_to_bit_type(t) and can_convert_to_bit_type(T):
                # If they can all be converted to Bit, then that's okay
                convert_to_bit = True
            else:
                raise ValueError(
                    "All fields in a Array or a Tuple must be the same type"
                    f"got {t} expected {T}")

    assert len(Ts)

    if convert_to_bit is True:
        ts = [bit(t) for t in ts]
        T = {
            Direction.In: In,
            Direction.Out: Out,
            Direction.InOut: InOut
        }[T.direction](Bit)

    value = totype[len(Ts), T](ts)
    if n is not None and len(value) < n:
        value = value.ext_to(n)
    return value


def array(value, n=None):
    return convertbits(value, n, Array, False)


def bits(value, n=None):
    return convertbits(value, n, Bits, True)


def uint(value, n=None):
    return convertbits(value, n, UInt, True)


def sint(value, n=None):
    return convertbits(value, n, SInt, True)


def bfloat(value, n=None):
    return convertbits(value, n, BFloat, True)


def concat(*arrays):
    ts = []
    for a in arrays:
        if isinstance(a, ht.BitVector):
            ts.extend(a.bits())
        elif isinstance(magma_value(a), Bit):
            ts.extend([a])
        elif isinstance(a, (bool, ht.Bit)):
            ts.extend([Bit(a)])
        else:
            if not isinstance(a, Array):
                raise TypeError(
                    "concat expects values of type Array, BitVector, Bit, or "
                    f"bool, not {a} with type {type(a)}"
                )
            ts.extend(a.ts)
    return array(ts)


def repeat(value, n):
    if isinstance(magma_value(value), Bit):
        repeats = bits(n * [value])
    else:
        repeats = array(n * [value])
    return repeats


def check_value_is_not_input(fn):
    @functools.wraps(fn)
    def wrapped(value, n):
        if isinstance(value, Type) and not value.is_output():
            raise Exception(f"{fn.__name__} only works with non input values")
        return fn(value, n)
    return wrapped


def zext(value, n):
    """Extend `value` by `n` zeros"""
    assert (isinstance(value, (UInt, SInt, Bits)) or
            (isinstance(value, Array) and issubclass(value.T, Digital)))
    if not is_int(n) or n < 0:
        raise TypeError(f"Expected non-negative integer, got '{n}'")
    if n == 0:
        return value
    if isinstance(value, UInt):
        zeros = uint(0, n)
    elif isinstance(value, SInt):
        zeros = sint(0, n)
    elif isinstance(value, Bits):
        zeros = bits(0, n)
    elif isinstance(value, Array):
        zeros = array(0, n)
    result = concat(value, zeros)
    if isinstance(value, UInt):
        return uint(result)
    elif isinstance(value, SInt):
        return sint(result)
    elif isinstance(value, Bits):
        return bits(result)
    return result


def zext_by(value, n):
    """Extend `value` by `n` zeros"""
    return zext(value, n)


def zext_to(value, n):
    """Extend `value` to length `n` with zeros"""
    return zext(value, n - len(value))


def sext(value, n):
    """Extend `value` by `n` replications of the msb (`value[-1]`)"""
    if not isinstance(value, SInt):
        raise TypeError(f"Expeted SInt, got {type(value).undirected_t}")
    if not is_int(n) or n < 0:
        raise TypeError(f"Expected non-negative integer, got '{n}'")
    if n == 0:
        return value
    return sint(concat(array(value), array([value[-1]] * n)))


def sext_by(value, n):
    """Extend `value` by `n` replications of the msb (`value[-1]`)"""
    return sext(value, n)


def sext_to(value, n):
    """Extend `value` to length `n` by replicating the msb (`value[-1]`)"""
    return sext(value, n - len(value))


from .bitutils import int2seq
from .array import Array
from .bit import Digital, Bit


#
# convert value to a tuple
#   *value = tuple from positional arguments
#   **kwargs = tuple from keyword arguments
#
def _tuple(value, n=None, t=Tuple):
    if isinstance(value, t):
        return value

    if not isinstance(magma_value(value),
                      (Digital, Array, IntegerTypes, Sequence, Mapping)):
        raise ValueError(
            "bit can only be used on a Bit, an Array, or an int; not {}".format(type(value)))

    decl = OrderedDict()
    args = []

    if isinstance(value, IntegerTypes):
        if n is None:
            n = max(value.bit_length(),1)
        value = int2seq(value, n)
    elif isinstance(magma_value(value), Digital):
        value = [value]
    elif isinstance(value, Array):
        value = [value[i] for i in range(len(value))]

    if isinstance(value, Sequence):
        ts = list(value)
        for i in range(len(ts)):
            args.append(ts[i])
            decl[i] = type(ts[i])
    elif isinstance(value, Mapping):
        for k, v in value.items():
            args.append(v)
            decl[k] = type(v)
    for a, d in zip(args, decl):
        # bool types to Bit
        if issubclass(decl[d], (bool, ht.Bit)):
            decl[d] = Digital
        # Promote integer types to Bits
        elif decl[d] in IntegerTypes:
            decl[d] = Bits[max(a.bit_length(), 1)]
        elif issubclass(decl[d], ht.BitVector):
            decl[d] = Bits[len(decl[d])]

    if t == Tuple:
        return t[tuple(decl.values())](*args)
    assert t == Product
    return t.from_fields("anon", decl)(*args)


def namedtuple(**kwargs):
    return _tuple(kwargs, t=Product)


def product(**kwargs):
    return _tuple(kwargs, t=Product)


def tuple_(value, n=None):
    return _tuple(value, n)


def replace(value, others: dict):
    value = magma_value(value)
    if isinstance(value, Product):
        d = dict(value.items())
        d.update(others)
        return namedtuple(**d)
    elif isinstance(value, Tuple):
        a = value.values()
        for idx, v in others.items():
            a[int(idx)] = v
        return tuple_(a)
    elif isinstance(value, Array):
        l = value.as_list()
        for idx, v in others.items():
            l[int(idx)] = v
        return array(l)
    else:
        raise ValueError("replace can only be used with an Array, a Tuple, or Product")


def as_bits(value):
    return bits(value.flatten())


def from_bits(cls, value):
    ts = value.ts
    return cls.unflatten(ts)



from collections import namedtuple, OrderedDict
from .circuit import CopyInstance, Circuit, IO
from .is_definition import isdefinition
from .is_primitive import isprimitive
from .digital import Digital
from .bit import *
from .clock import Clock, Enable, Reset, AsyncReset, wiredefaultclock
from .array import *
from .tuple import Tuple
from .wire import wire
from .conversions import array, tuple_
from .ref import DefnRef, InstRef, ArrayRef, TupleRef
from .scope import *
from magma.circuit import Circuit


__all__ = ['TransformedCircuit', 'flatten', 'setup_clocks', 'get_uniq_circuits']

class MagmaTransformException(Exception):
    pass


# Holds the new modified circuit as well as a map from old bits and scopes to
# new bits
class TransformedCircuit:
    def __init__(self, orig_circuit, transform_name):
        # Maps from original bits to bits in transformed circuit
        self.orig_to_new = {}

        decl = orig_circuit.interface.decl()
        ports = dict(zip(decl[::2], decl[1::2]))
        class _TransformedCircuit(Circuit):
            name = orig_circuit.name + '_' + transform_name
            io = IO(**ports)
        self.circuit = _TransformedCircuit

    def get_new_bit(self, orig_bit, scope):
        assert isinstance(scope, Scope), "Second argument to get_new_bit should be an instance of Scope"
        if isinstance(orig_bit, Array):
            arr = []
            for o in orig_bit:
                arr.append(self.get_new_bit(o, scope))

            arr = array(arr)
            if arr.iswhole():
                return arr[0].name.array

            return arr
        else:
            try:
                return self.orig_to_new[QualifiedBit(bit=orig_bit, scope=scope)]
            except KeyError:
                raise MagmaTransformException("Could not find bit in transform mapping. bit={}, scope={}".format(orig_bit, scope))

    def set_new_bit(self, orig_bit, orig_scope, new_bit):
        assert isinstance(new_bit, (Digital, Array, Tuple)), type(new_bit)

        if isinstance(orig_bit, (Array, Tuple)):
            # Map the individual bits
            for o, n in zip(orig_bit, new_bit):
                self.orig_to_new[QualifiedBit(bit=o, scope=orig_scope)] = n
        else:
            self.orig_to_new[QualifiedBit(bit=orig_bit, scope=orig_scope)] = new_bit

def get_primitives(outer_circuit, outer_scope):
    primitives = []
    mapping = {}

    for instance in outer_circuit.instances:
        inner_scope = Scope(parent=outer_scope, instance=instance)
        inner_circuit = type(instance)

        if isprimitive(inner_circuit):
            primitives.append(QualifiedInstance(instance=instance, scope=outer_scope))
        else:
            for name, outerbit in instance.interface.ports.items():
                innerbit = inner_circuit.interface.ports[name]
                outerloc = QualifiedBit(bit=outerbit, scope=outer_scope)
                innerloc = QualifiedBit(bit=innerbit, scope=inner_scope)

                if isinstance(outerbit, Array):
                    for o, i in zip(outerbit, innerbit):
                        oqual = QualifiedBit(bit=o, scope=outer_scope)
                        iqual = QualifiedBit(bit=i, scope=inner_scope)
                        if o.is_input():
                            mapping[iqual] = oqual
                        else:
                            mapping[oqual] = iqual

                if outerbit.is_input():
                    mapping[innerloc] = outerloc
                else:
                    mapping[outerloc] = innerloc

            instance_primitives, inst_map = get_primitives(inner_circuit, inner_scope)

            for k, v in inst_map.items():
                mapping[k] = v

            primitives += instance_primitives

    return (primitives, mapping)


def get_renamed_port(circ, name):
    for key, value in circ.renamed_ports.items():
        if value == name:
            return circ.interface.ports[key]
    return circ.interface.ports[name]

def get_new_source(source_qual, primitive_map, old_circuit, new_circuit):
    old_source = source_qual.bit
    scope = source_qual.scope

    if old_source.const():
        return old_source

    bitref = old_source.name
    idxs = []
    while isinstance(bitref, (ArrayRef, TupleRef)):
        idxs.append(bitref.index)
        if isinstance(bitref, ArrayRef):
            bitref = bitref.array.name
        else:
            bitref = bitref.tuple.name

    if isinstance(bitref, InstRef):
        # Get the primitive outbit is attached to,
        old_primitive = bitref.inst
        if not isprimitive(type(old_primitive)):
            raise MagmaTransformException("Failed to collapse bit to primitive. bit={} type={}".format(old_primitive, type(old_primitive)))
        new_primitive = primitive_map[QualifiedInstance(instance=old_primitive, scope=scope)]
        newsource = get_renamed_port(new_primitive, bitref.name)
    elif isinstance(bitref, DefnRef):
        defn = bitref.defn.name
        assert defn == old_circuit.name, f"Collapsed bit to circuit other than outermost, {defn} {old_circuit.name}"
        newsource = get_renamed_port(new_circuit, bitref.name)
    elif isinstance(old_source, (Array, Tuple)):
        # Must not be a whole array
        assert bitref.anon()

        new_src = []
        for s in old_source:
            partial_qual = QualifiedBit(bit=s, scope=scope)
            ns = get_new_source(partial_qual, primitive_map, old_circuit, new_circuit)
            new_src.append(ns)

        if isinstance(old_source, Array):
            return array(new_src)
        return tuple_(new_src)
    else:
        assert False, f"Failed to collapse bit {bitref}"

    if idxs:
        for idx in reversed(idxs):
            newsource = newsource[idx]

    return newsource

def wire_new_bit(origbit, newbit, cur_scope, primitive_map, bit_map, old_circuit, flattened_circuit):
    if isinstance(origbit, (Array, Tuple)):
        assert isinstance(newbit, type(origbit))
        for x, y in zip(origbit, newbit):
            wire_new_bit(x, y, cur_scope, primitive_map, bit_map, old_circuit, flattened_circuit)
        return
    # Trace back from origbit to its source, tracking bits that will be "collapsed"
    new_circuit = flattened_circuit.circuit

    sourcebit = origbit.trace()
    if sourcebit is None:
        if isinstance(origbit, Array):
            # TODO: Raise an exception for now, can we handle this case silently (ignore unwired ports)?
            raise MagmaTransformException("Calling `.value()` on Array returned None. Array = {}, values = {}. Likely an unwired port.".format(origbit, [b.value() for b in origbit.ts]))
        else:
            raise MagmaTransformException("Calling `.value()` on {} returned None. Likely an unwired port.".format(origbit))
    source_qual = QualifiedBit(bit=sourcebit, scope=cur_scope)
    orig_qual = QualifiedBit(bit=origbit, scope=cur_scope)
    collapsed_out_bits = [source_qual]
    collapsed_in_bits = [orig_qual]

    while source_qual in bit_map:
        intermediate_in = bit_map[source_qual]
        sourcebit = intermediate_in.bit.value()
        if sourcebit is None:
            raise Exception("Sourcebit is None, {} is possibly unwired?".format(source_qual))

        source_qual = QualifiedBit(bit=sourcebit, scope=intermediate_in.scope)
        collapsed_out_bits.append(source_qual)
        collapsed_in_bits.append(intermediate_in)

    newsource = get_new_source(source_qual, primitive_map, old_circuit, new_circuit)
    wire(newsource, newbit)

    for collapsed in collapsed_in_bits:
        flattened_circuit.set_new_bit(collapsed.bit, collapsed.scope, newbit)

    for collapsed in collapsed_out_bits:
        flattened_circuit.set_new_bit(collapsed.bit, collapsed.scope, newsource)

def flatten(circuit):
    flattened_circuit = TransformedCircuit(circuit, 'flattened')

    outer_scope = Scope()
    orig_primitives, bit_map = get_primitives(circuit, outer_scope)

    primitive_map = {} # Maps from primitives in the old circuit to new circuits
    new_primitives = []
    with flattened_circuit.circuit.open():
        for old in orig_primitives:
            new = CopyInstance(old.instance)
            new_primitives.append(new)
            primitive_map[old] = new

        # Wire up all the new instances
        for new_inst, qual_inst in zip(new_primitives, orig_primitives):
            orig_inst = qual_inst.instance

            for name, origbit in orig_inst.interface.ports.items():
                if origbit.is_input():
                    newbit = new_inst.interface.ports[name]
                    wire_new_bit(origbit, newbit, qual_inst.scope, primitive_map, bit_map, circuit, flattened_circuit)

        # Finally, wire up the circuit outputs
        new_circuit = flattened_circuit.circuit
        for name, origbit in circuit.interface.ports.items():
            if origbit.is_input(): # Circuit output
                newbit = new_circuit.interface.ports[name]
                wire_new_bit(origbit, newbit, Scope(), primitive_map, bit_map, circuit, flattened_circuit)

        # Handle unwired inputs
        for name, origbit in circuit.interface.ports.items():
            if origbit.is_output() and not origbit.wired():
                newbit = new_circuit.interface.ports[name]
                flattened_circuit.set_new_bit(origbit, Scope(), newbit)

    return flattened_circuit



from abc import ABC, abstractmethod
import ast
from collections import Counter
import inspect
import textwrap
from .config import get_debug_mode
from .ref import LazyCircuit
from .view import InstView


def _parse_name(inst):
    """
    Returns 'name' if the source where @inst is instanced looks like:

      <name> = <type>(...)

    otherwise, returns None.
    """
    try:
        # File may not exist (e.g. in sequential we exec a string)
        with open(inst.debug_info.filename, "r") as f:
            line = f.read().splitlines()[inst.debug_info.lineno - 1]
            tree = ast.parse(textwrap.dedent(line)).body[0]
    except Exception:
        # Sometimes we cannot parse the line, e.g.
        # x = [Circuit()
        #      for _ in range(8)]
        # Will give use the second line for the list comprehension, which
        # cannot be parsed.  It's probably generally difficult to catch all
        # these cases
        # For now, we just silently ignore exceptions and just assume we can't
        # find a good debug name
        return None
    # Simple case when <Name> = <Instance>().
    if isinstance(tree, ast.Assign) and len(tree.targets) == 1 \
       and isinstance(tree.targets[0], ast.Name):
        return tree.targets[0].id
    return None


class PlacerBase(ABC):
    @abstractmethod
    def place(self, inst):
        raise NotImplementedError()

    @abstractmethod
    def instances(self):
        raise NotImplementedError()


def _setup_view(inst):
    # Setup view now because inline strings might use it during defn
    for sub_inst in getattr(type(inst), "instances", []):
        setattr(inst, sub_inst.name, InstView(sub_inst, InstView(inst)))


class Placer:
    def __init__(self, defn):
        self._defn = defn
        self._name_counter = Counter()
        self._instance_counter = Counter()
        self._instance_name_map = {}
        self._instances = []
        self._finalized = False

    def instances(self):
        return self._instances

    def _user_name(self, inst) -> bool:
        """
        Returns True if @inst already has a name, False otherwise. Does not
        modify @inst in either case.
        """
        if not inst.name:
            return False
        self._name_counter[inst.name] += 1
        return True

    def _debug_name(self, inst) -> bool:
        """
        Tries to retrieve a debug name from the source of the instance; sets the
        name of @inst and returns True if available. Otherwise (or if debug mode
        is disabled), returns False.
        """
        if not get_debug_mode():
            return False
        basename = _parse_name(inst)
        if not basename:
            return False
        count = self._name_counter[basename]
        inst.name = basename if count == 0 else f"{basename}_{count}"
        # If this is the first duplicate of basename we've seen (e.g. in a
        # second-iteration of a loop), then we need to rename the first instance
        # with a appended '_0'.
        if count == 1:
            first = self._instance_name_map[basename]
            first.name += "_0"
            del self._instance_name_map[basename]
            self._instance_name_map[first.name] = first
        self._name_counter[basename] += 1
        return True

    def _definition_name(self, inst) -> bool:
        """
        Sets name of @inst based on the type of @inst, of the form
        {type.name}_inst{count}. Always returns True.
        """
        type_name = type(inst).name
        count = self._instance_counter[type_name]
        inst.name = f"{type_name}_inst{count}"
        self._name_counter[inst.name] += 1
        self._instance_counter[type_name] += 1
        return True

    def _name(self, inst):
        # Sets the name of @inst by delegating to methods _user_name,
        # _debug_name, _definition_name. Note that because the 'or' is
        # short-circuited, we call as many of these as needed until one returns
        # True. We expect exactly one to return True, and assert that.
        assert (self._user_name(inst)
                or self._debug_name(inst)
                or self._definition_name(inst))
        self._instance_name_map[inst.name] = inst

    def place(self, inst):
        self._name(inst)
        self._instances.append(inst)
        inst.defn = self._defn
        if get_debug_mode():
            inst.stack = inspect.stack()
        _setup_view(inst)

    def finalize(self, defn):
        if self._finalized:
            raise Exception("Can only call finalize on a placer once")
        self._finalized = True
        return self


class StagedPlacer(ABC):
    def __init__(self, name):
        self._name = name
        self._instances = []
        self._finalized = False

    @property
    def instances(self):
        return self._instances.copy()

    @property
    def name(self):
        return self._name

    def place(self, inst):
        self._instances.append(inst)
        inst.defn = LazyCircuit
        _setup_view(inst)

    def finalize(self, defn):
        if self._finalized:
            raise Exception("Can only call finalize on a staged placer once")
        placer = Placer(defn)
        for inst in self._instances:
            placer.place(inst)
        self._finalized = True
        return placer



import enum

from magma.passes.passes import CircuitPass, pass_lambda


_HAS_ERROR_KEY = "_has_errors_"


def _has_errors(ckt):
    return getattr(ckt, _HAS_ERROR_KEY, False)


class FindErrorsPass(CircuitPass):

    class ErrorReportingMode(enum.Enum):
        FIRST = enum.auto()
        ALL = enum.auto()
        COLLECT = enum.auto()

    def __init__(self, main,
                 mode: ErrorReportingMode = ErrorReportingMode.FIRST):
        super().__init__(main)
        self._mode = mode
        self._errors = []

    def __call__(self, ckt):
        if not _has_errors(ckt):
            return
        self._handle_error(ckt)

    @property
    def errors(self):
        if self._mode == FindErrorsPass.ErrorReportingMode.FIRST:
            return None
        collect = (self._mode == FindErrorsPass.ErrorReportingMode.ALL or
                   self._mode == FindErrorsPass.ErrorReportingMode.COLLECT)
        if collect:
            return self._errors.copy()
        raise NotImplementedError(self._mode)

    def _handle_error(self, ckt):
        if self._mode == FindErrorsPass.ErrorReportingMode.FIRST:
            assert not self._errors
            raise Exception(f"Found circuit with errors: {ckt.name}")
        collect = (self._mode == FindErrorsPass.ErrorReportingMode.ALL or
                   self._mode == FindErrorsPass.ErrorReportingMode.COLLECT)
        if collect:
            self._errors.append(ckt)
            return
        raise NotImplementedError(self._mdoe)

    def done(self):
        # If mode == ALL, then raise an error with all the circuit names. In the
        # case of COLLECT or FIRST, do nothing.
        if self._mode == FindErrorsPass.ErrorReportingMode.ALL:
            msg = (f"Found circuits with errors: "
                   f"{ckt.name for ckt in self._errors}")
            raise Exception(msg)


find_errors_pass = pass_lambda(FindErrorsPass)



from .passes import EditDefinitionPass
from ..is_definition import isdefinition
from magma.array import Array
from magma.clock import (wire_default_clock, is_clock_or_nested_clock,
                         get_default_clocks)
from magma.tuple import Tuple


def _drive_undriven_children(port, clocks):
    # list comp so it doesn't short circuit
    undrivens = [_drive_if_undriven_input(p, clocks) for p in port]
    return any(undrivens)


def _drive_if_undriven_input(port, clocks):
    if port.is_mixed():
        return _drive_undriven_children(port, clocks)
    if port.is_input() and port.trace() is None:
        if (isinstance(port, (Array, Tuple)) and
                any(t.trace() is not None for t in port)):
            # Partially undriven, recurse through to drive only the undriven
            # children
            return _drive_undriven_children(port, clocks)
        if (not is_clock_or_nested_clock(type(port)) or
                not wire_default_clock(port, clocks)):
            port.undriven()
        # We always return True, even if we do default clock wiring since in
        # this case, it should be a definition now
        return True
    return False


def _drive_undriven(interface, clocks):
    undriven = False
    for port in interface.ports.values():
        undriven |= _drive_if_undriven_input(port, clocks)
    return undriven


class DriveUndrivenPass(EditDefinitionPass):
    def edit(self, circuit):
        clocks = get_default_clocks(circuit)
        if _drive_undriven(circuit.interface, clocks):
            circuit._is_definition = True
        if not isdefinition(circuit):
            return
        for inst in circuit.instances:
            _drive_undriven(inst.interface, clocks)
